/*-
 * Copyright 2021 Hochschule für Technik Stuttgart
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package de.hft.stuttgart.citygml.viewer;

import java.nio.FloatBuffer;

import org.joml.Math;
import org.joml.Matrix4f;
import org.joml.Vector2f;
import org.joml.Vector3f;
import org.lwjgl.system.MemoryUtil;

public class Camera {

	private static final float MIN_DISTANCE = 1f;
	private static final float PI = (float) Math.PI;

	Matrix4f projMatrix;
	private Matrix4f viewMatrix;
	private Matrix4f projViewMatrix;

	private float rotateAroundZ = 0f;

	private Vector3f eyePos = new Vector3f();
	private Vector3f centerPos = new Vector3f();
	private Vector3f up = new Vector3f();
	private float distance = 5;

	private static final float SCROLL_PERCENTAGE = 0.1f;

	UniformLocation uniformLocation;
	FloatBuffer matrixBuffer = MemoryUtil.memAllocFloat(16);

	public Camera(Shader shader) {
		uniformLocation = shader.getUniformLocation("projViewModel");
		projViewMatrix = new Matrix4f();
		viewMatrix = new Matrix4f();
		projMatrix = new Matrix4f();
		up.z = 1;
		eyePos.x = distance;
	}

	public void reshape(int width, int height, float distance) {
		// 90 degree fow
		float fow = PI / 2f;
		float aspectRatio = (float) width / height;
		projMatrix.setPerspective(fow, aspectRatio, MIN_DISTANCE, distance);
		updateMatrix();
	}

	private void updateMatrix() {
		viewMatrix.setLookAt(eyePos, centerPos, up);
		projMatrix.mul(viewMatrix, projViewMatrix);
		// model matrix is identity
		projViewMatrix.get(matrixBuffer);
		uniformLocation.uploadMat4(matrixBuffer);
	}

	public void rotate(double dragDiffX, double dragDiffY) {
		float rotationX = (float) (-dragDiffX / 500f);
		float rotationZ = (float) (dragDiffY / 500f);
		float tempRotationValue = rotateAroundZ + rotationZ;
		if (tempRotationValue < -Math.PI / 2 + 0.0001 || tempRotationValue > Math.PI / 2 - 0.0001) {
			// to close to 90 degree, stop rotation
			rotationZ = 0;
		}
		rotateAroundZ += rotationZ;

		Vector3f res = new Vector3f();
		eyePos.sub(centerPos, res);
		res.rotateZ(rotationX);

		Vector3f rotAxis = new Vector3f(res.x, res.y, res.z + 5);
		rotAxis.cross(res);
		rotAxis.normalize();
		res.rotateAxis(rotationZ, rotAxis.x, rotAxis.y, rotAxis.z);
		centerPos.add(res, eyePos);
		updateMatrix();
	}

	public void setDistance(float distance) {
		this.distance = distance;
		Vector3f res = new Vector3f();
		eyePos.sub(centerPos, res);
		res.normalize().mul(distance);
		centerPos.add(res, eyePos);
		updateMatrix();
	}

	public void scroll(int scroll) {
		int addedDistance = (int) (distance * scroll * SCROLL_PERCENTAGE);
		distance += addedDistance;
		setDistance(distance);
	}

	public void free() {
		MemoryUtil.memFree(matrixBuffer);
	}

	public void move(double dragDiffX, double dragDiffY) {
		Vector3f res = new Vector3f();
		centerPos.sub(eyePos, res);
		Vector2f dir = new Vector2f(res.x, res.y);
		dir.normalize();
		Vector2f yDrag = new Vector2f();
		dir.mul((float) -dragDiffY * 0.5f, yDrag);

		// handle diffY
		centerPos.add(yDrag.x, yDrag.y, 0);
		eyePos.add(yDrag.x, yDrag.y, 0);

		// handle diffX
		float temp = dir.x;
		dir.x = dir.y;
		dir.y = -temp;
		dir.mul((float) dragDiffX * 0.5f);

		centerPos.add(dir.x, dir.y, 0);
		eyePos.add(dir.x, dir.y, 0);

		updateMatrix();
	}

}
